/*
 * Copyright 2020 Confluent Inc.
 *
 * Licensed under the Confluent Community License (the "License"); you may not use
 * this file except in compliance with the License.  You may obtain a copy of the
 * License at
 *
 * http://www.confluent.io/confluent-community-license
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 */

package io.confluent.ksql.execution.expression.tree;

import io.confluent.ksql.name.ColumnName;
import io.confluent.ksql.name.SourceName;
import io.confluent.ksql.parser.NodeLocation;
import java.util.Objects;
import java.util.Optional;

public abstract class ColumnReferenceExp extends Expression {

  final ColumnName name;

  ColumnReferenceExp(
      final Optional<NodeLocation> location,
      final ColumnName name
  ) {
    super(location);
    this.name = Objects.requireNonNull(name, "name");
  }

  public ColumnName getColumnName() {
    return name;
  }

  public abstract Optional<SourceName> maybeQualifier();
}
