/*
 * Copyright 2020 Confluent Inc.
 *
 * Licensed under the Confluent Community License (the "License"); you may not use
 * this file except in compliance with the License.  You may obtain a copy of the
 * License at
 *
 * http://www.confluent.io/confluent-community-license
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 */

package io.confluent.ksql.schema.ksql;

/**
 * Helpers for working with Sql {@code DOUBLE}.
 */
public final class SqlDoubles {

  private SqlDoubles() {
  }

  /**
   * Parse a SQL double from a string.
   *
   * <p>Rejects {@code Infinity} and {@code Nan} as invalid.
   *
   * @param str the string to parse.
   * @return the double value.
   */
  public static double parseDouble(final String str) {
    final double d = Double.parseDouble(str);
    if (Double.isInfinite(d) || Double.isNaN(d)) {
      throw new NumberFormatException("Invalid double value: " + str);
    }
    return d;
  }
}
